/*****************************************************************************
 *                Copyright (C) 2013 Radian Memory Systems
 *
 * This file is provided under non-disclosure agreement.
 * See the file COPYING for details.
 ******************************************************************************/
/* dialog_poll.c
 *
 * Utility to get notification and dialog page log from the kernel using
 * the genetlink mechanism
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <poll.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <signal.h>
#include <netlink/netlink.h>
#include <netlink/genl/genl.h>
#include <netlink/genl/ctrl.h>
#include <linux/genetlink.h>
typedef __u64 u64;
typedef __u32 u32;
typedef __u16 u16;
typedef __u8 u8;
#include "dialogDef.h"
#include "dialog_util.h"

#define	NVME_EVENT_NAME		"nvme_event"
#define NVME_EVENT_GROUP_NAME	"nvme_mc_group"
/*
 * This routine replaces the standard net link message parsing
 * in order to dump the dialog page
*/
static int
parse_cb(
	struct nl_msg *msg,
	void *arg)
{
	struct nlmsghdr *nlh = nlmsg_hdr(msg);
	struct nlattr *attrs[NLA_TYPE_MAX];
	DialogPage *dialogPtr;
	u32	   dialogPageNum;
	int ret;
	FILE *fd = (FILE *) arg;

	if (fd == NULL)
		fd = stdout;
	memset(attrs, 0 , sizeof(attrs));

	/* Validate message and parse attributes */
	ret = genlmsg_parse(nlh, 0, attrs, NLA_TYPE_MAX, NULL);

	if ((ret == 0) && (attrs[NLA_STRING] != NULL)) {
		/* Skip beyond the attribute structure */
		dialogPageNum = *(((u32 *)(attrs[NLA_STRING])) + 1);
		if (dialogPageNum == GLP_ID_DIALOG) {
			dialogPtr = (DialogPage *)
					((u32 *)attrs[NLA_STRING] + 3);
			output_dialog(dialogPtr, fd);
		} else if (dialogPageNum != 1) {
			fprintf(fd,
				"Wrong dialog page %08x\n",
				dialogPageNum);
		}
	} else {
		fprintf(fd,
			"Problem parsing generic ret %d attr %p\n",
			ret,
			attrs[NLA_STRING]);
	}
	fflush(fd);

	return 0;
}

static void
wait_for_q(void)
{
	system("/bin/stty raw");
	getchar();
	system("/bin/stty cooked");
}

static void
description(void)
{
	printf("This command waits for a kernel event that posts\n");
	printf("the dialog page from any RMS-200\n");
	return;
}

static void
usage(void)
{
	printf("usage: dialog_poll [-l] [-w logfile]\n");
	printf("  -l	  : listen forever\n");
	printf("  -w file : dumps the contents into a file\n");
}

int
main(int argc, char *argv[])
{
	int c;
	FILE *fd = 0;
	int write_file = 0;
	char *logfile = NULL;
	int loop = 0;
	struct nl_sock *sock;
	int family, group;

	while (1) {
		c = getopt(argc, argv, "w:lhH");
		if (c < 0)
			break;

		switch (c) {
		case 'w':
			logfile = strdup(optarg);
			printf("write to file %s\n", logfile);
			write_file = 1;
			break;
		case 'l':
			printf("listen forever\n");
			loop = 1;
			break;
		case 'H':
			description();
			usage();
			wait_for_q();
			exit(1);
			break;
		case 'h':
		default:
			usage();
			exit(1);
		}
	}

	if (write_file) {
		fd = fopen(logfile, "a+");
		if (fd == NULL) {
			perror("Cannot open output file\n");
			exit(1);
		}
	}
	/* Allocate a new netlink socket */
	sock = nl_socket_alloc();

	nl_join_groups(sock, 1);

	/* Connect to generic netlink socket on kernel side */
	genl_connect(sock);

	/* Ask kernel to resolve family name to family id */
	family = genl_ctrl_resolve(sock, NVME_EVENT_NAME);
	if (family < 0) {
		printf("Bad family %d\n", family);
		nl_socket_free(sock);
		if (fd)
			fclose(fd);
		return 0;
	}

	group = genl_ctrl_resolve_grp(sock,
					NVME_EVENT_NAME,
					NVME_EVENT_GROUP_NAME);

	if (group < 0 || family < 0) {
		printf("Bad multicast group %d\n", group);
		nl_socket_free(sock);
		if (fd)
			fclose(fd);
		return 0;
	}

	if (!nl_socket_add_membership(sock, group)) {
		/*
		 * Prepare socket to receive the answer by specifying the
		 * callback function to be called for valid messages.
		 */
		nl_socket_modify_cb(sock,
					NL_CB_MSG_IN,
					NL_CB_CUSTOM,
					parse_cb,
					fd);

		/* Wait for the answer and receive it */
		do {
			nl_recvmsgs_default(sock);
		} while (loop == 1);

	} else {
		printf("Error adding membership %d\n", group);
	}

	nl_socket_free(sock);
	if (fd)
		fclose(fd);

	return 0;
}


